/*
Copyright  2006-2009 Eland Systems All Rights Reserved.

   1. Redistribution and use in source and binary forms must retain the above
   copyright notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by Eland Systems

   4. The name of Eland Systems may not be used to endorse or promote products
   derived from this software without specific prior written permission.

   Author scam+archiver@elandsys.com
*/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sysexits.h>
#include <unistd.h>
#include <errno.h>
#include <stdarg.h>
#include <ctype.h>
#include <pwd.h>
#include <grp.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <syslog.h>
#include <pthread.h>
#include "libmilter/mfapi.h"

#include "util.h"

#define PIDFILE "/var/spool/scam/scam-archiver.pid"
#define SCAMCONF "/etc/mail/scam.conf"
#define MLFIPRIV	((struct mlfiPriv *) smfi_getpriv(ctx))

#define VERSION "1.0"

static int testmode = 0;
char *archivaddr = NULL;

struct mlfiPriv {
	char	*queueid;
} mlfi_priv;


static sfsistat mlfi_cleanup(SMFICTX *);

static sfsistat
mlfi_connect(SMFICTX *ctx, char *hostname,  _SOCK_ADDR *hostaddr)
{
	struct mlfiPriv *priv= MLFIPRIV;
	struct sockaddr_in  *remoteaddr;
	struct in_addr addr;


	priv = malloc(sizeof *priv);
	if (priv == NULL) {
		/* can't accept this message right now */
		syslog (LOG_ERR, "cannot allocate priv");
		return SMFIS_TEMPFAIL;
	}
	memset(priv, '\0', sizeof *priv);

	/* set the private data */
	if (smfi_setpriv(ctx, priv) != MI_SUCCESS)
	{
		smfi_setpriv(ctx, priv);
		free(priv);

		syslog (LOG_ERR, "cannot setpriv");
		(void) mlfi_cleanup(ctx);
        return SMFIS_TEMPFAIL;
	}

	priv->queueid = NULL;

	return SMFIS_CONTINUE;
}


static sfsistat
mlfi_envfrom(SMFICTX *ctx, char **argv)
{
	struct mlfiPriv *priv= MLFIPRIV;
	char *authtype;
	char *queueid;

	queueid= smfi_getsymval(ctx, "i");
	if (queueid != NULL)
	{
		if (priv->queueid != NULL)
			free(priv->queueid);

		if ((priv->queueid = strdup (queueid)) == NULL)
		{
			syslog (LOG_ERR, "cannot allocate queueid");
			(void) mlfi_cleanup(ctx);
			return SMFIS_TEMPFAIL;
		}
	}

	return SMFIS_CONTINUE;

}

static sfsistat
mlfi_envrcpt(SMFICTX *ctx, char **argv)
{
	return SMFIS_CONTINUE;
}



static sfsistat
mlfi_eom(SMFICTX *ctx)
{
	if (testmode == 0)
	{
		smfi_addrcpt( ctx, archivaddr);
	}
#ifdef SM812
	smfi_addheader(ctx, "X-Scam-Archiver", "www.elandsys.com/scam/");
#else
	smfi_insheader(ctx, 1, "X-Scam-Archiver", "www.elandsys.com/scam/");
#endif
	return SMFIS_CONTINUE;
}

sfsistat
mlfi_abort(SMFICTX *ctx)
{

	mlfi_cleanup(ctx);

	return SMFIS_CONTINUE;
}

static sfsistat
mlfi_cleanup(SMFICTX *ctx)
{
	struct mlfiPriv *priv = MLFIPRIV;

	if (priv == NULL)
		return;

	if (priv->queueid)
	{
		free(priv->queueid);
		priv->queueid = NULL;
	}

	return;

}

static sfsistat
mlfi_close(SMFICTX *ctx)
{
	struct mlfiPriv *priv = MLFIPRIV;

	mlfi_cleanup(ctx);


	if (priv)
	{
		free(priv);
	}
	smfi_setpriv(ctx, NULL);

	return SMFIS_CONTINUE;
}

struct smfiDesc smfilter =
{
	"scam-archiver",	/* filter name */
	SMFI_VERSION,	/* version code -- do not change */
	SMFIF_ADDHDRS|SMFIF_ADDRCPT,	/* flags */
	mlfi_connect,		/* connection info filter */
	NULL,		/* SMTP HELO command filter */
	mlfi_envfrom,	/* envelope sender filter */
	mlfi_envrcpt,	/* envelope recipient filter */
	NULL,				/* header filter */
	NULL,	/* end of header */
	NULL,			/* body block filter */
	mlfi_eom,	/* end of message */
	mlfi_abort,	/* message aborted */
	mlfi_close,	/* connection cleanup */
};

static void
usage()
{
	fprintf(stderr, "scam-archiver version %s\n", VERSION);
    fprintf(stderr, "usage: scam-archiver -p protocol:address [-T timeout] [-R] [-t] [-D]\n");
	fprintf(stderr, "          -D run as a daemon\n");
	fprintf(stderr, "          -t run in test mode\n");
	fprintf(stderr, "          -u run as specified user\n");
}

static void
catch_signal( int signo )
{
	switch (signo)
	{
		case SIGUSR1:
		   syslog (LOG_NOTICE, "Caught SIGUSR1, exiting...");
			smfi_stop();
			sleep(10);
			exit (0) ;
	}

}

int
write_pid()
{
	char buf[20];
	int fd;
	long pid;

	if ((fd = open (PIDFILE, O_CREAT | O_TRUNC | O_WRONLY, 0600)) == -1)
	{
		syslog (LOG_ERR, "cannot open pidfile ");
	} else {
		pid = getpid ();
		snprintf (buf, sizeof (buf), "%ld", (long) pid);
		if (write (fd, buf, strlen (buf)) != strlen (buf)) {
			syslog (LOG_INFO, "cannot write to pidfile ");
		}
		close(fd);
	}
	return pid;
}

void
daemonize()
{

	signal (SIGUSR1, catch_signal);

	(void)close(0);
	(void)open("/dev/null", O_RDONLY, 0);
	(void)close(1);
	(void)open("/dev/null", O_RDONLY, 0);
	(void)close(2);
	(void)open("/dev/null", O_RDONLY, 0);

	switch (fork ())
	{
		case 0:
			syslog (LOG_INFO, "scam-archiver running ");
			break;

		case -1:
			perror ("cannot fork\n");
			exit (3);
			break;

		default:
			exit(0);
	}

	if (setsid() < 0)
		exit(1);

	write_pid();
	chdir("/");

}

int
read_conf()
{
	FILE *fh;
	const LINELENGTH = 1024;
	char line[LINELENGTH + 1];
	char buf[LINELENGTH + 1];
	char bufstr[LINELENGTH + 1];
	unsigned int num;
	unsigned long l;

	if ((fh = fopen( SCAMCONF, "r")) == NULL) {
		return 1;
	}

	syslog (LOG_INFO, "reading configuration");

	while( fgets(line, LINELENGTH, fh) )
	{
		if (line[0] == '#')
			continue;

		if (sscanf( line, "ArchiverAddr=%255[a-zA-Z0-9@+-.,]", buf) == 1)
		{
			if (strlen(buf) > 1)
			{
				archivaddr = strdup(buf);
				syslog (LOG_DEBUG, "ArchiverAddr %s", buf);
			}
		}
	}
	fclose(fh);


	if (archivaddr == NULL)
	{
		syslog (LOG_INFO, "ArchiverAddr required");
		exit(1);
	}

	return 0;
}

int
main(int argc, char *argv[])
{
	int c;
	int ret;
	const char *args = "p:T:u:g::thD";
	extern char *optarg;
	int daemonmode = 0;
	struct passwd *passwd = NULL;
	struct group *grp;
	char *user = NULL;
	char *group = NULL;
	uid_t uid;
	gid_t gid;

	umask(077);

	/* already a daemon */
	if(getppid()==1)
		return 0;

	/* Process command line options */
	while ((c = getopt(argc, argv, args)) != (char)EOF)
	{
		switch (c)
		{
		  case 'p':
			if (optarg == NULL || *optarg == '\0')
			{
				(void) fprintf(stderr, "Illegal connection: %s\n", optarg);
				exit(EX_USAGE);
			}
			if (smfi_setconn(optarg) == MI_FAILURE)
			{
				(void) fprintf(stderr, "smfi_setconn failed");
				exit(EX_SOFTWARE);
			}

			/*
				** If we're using a local socket, make sure it doesn't
				** already exist.
			*/
            if(strncmp(optarg, "unix:", 5) == 0)
                unlink(optarg + 5);
            else if(strncmp(optarg, "local:", 6) == 0)
                unlink(optarg + 6);

			break;

		  case 'T':
            if (optarg == NULL || *optarg == '\0')
            {
                (void) fprintf(stderr, "Illegal timeout: %s\n", optarg);
                exit(EX_USAGE);
            }
            if(smfi_settimeout(atoi(optarg)) == MI_FAILURE)
            {
                (void) fputs("smfi_settimeout failed", stderr);
                exit(EX_SOFTWARE);
            }
            break;

			case 'u':
				if (optarg == NULL || *optarg == '\0')
				{
					(void) fprintf(stderr, "Invalid username: %s\n", optarg);
					exit(EX_USAGE);
				}
				user = strdup(optarg);
				if ((passwd = getpwnam(user)) != NULL)
				{
					uid = passwd->pw_uid;
					gid = passwd->pw_gid;
				}
				break;

			case 'g':
				if (optarg == NULL || *optarg == '\0')
				{
					(void) fprintf(stderr, "Invalid group name: %s\n", optarg);
					exit(EX_USAGE);
				}
				group = strdup(optarg);
				if ((grp = getgrnam(group)) != NULL)
				{
					gid = grp->gr_gid;
				}
				break;

		  case 'd':
			if (optarg == NULL || *optarg == '\0')
            {
                (void) fprintf(stderr, "Illegal debug value: %s\n", optarg);
                exit(EX_USAGE);
            }
			if (smfi_setdbg(atoi(optarg)) == MI_FAILURE)
				fprintf(stderr, "smfi_setdbg enabled\n");
			break;

		  case 'D':
			  daemonmode = 1;
			  break;

		  case 't':
			  testmode = 1;
			  break;

		  case 'h':
		  default:
			usage();
			exit(0);

		}
	}

	(void) closelog();
	openlog("scam-archiver", LOG_PID | LOG_NDELAY, LOG_MAIL);

	setgid(gid);
	setuid(uid);

	if (!getuid() || !geteuid())
	{
		syslog (LOG_ERR, "scam-archiver cannot run as root");
		exit(0);
	}

	if (smfi_register(smfilter) == MI_FAILURE)
	{
		fprintf(stderr, "smfi_register failed\n");
		syslog (LOG_ERR, "smfi_register failed");
		exit(EX_UNAVAILABLE);
	}

	read_conf();

	if (daemonmode == 1)
	{
		daemonize();
	}

	ret = smfi_main();
	unlink(PIDFILE);
	syslog (LOG_INFO, "Exit");
	return(ret);
}

